package nl.uvt.locator;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.ParsingException;
import nu.xom.Text;
import nu.xom.ValidityException;
import nu.xom.XPathContext;

import org.apache.log4j.*;

public class Locator extends HttpServlet {

	public static final String SOAPNamespace   = "http://schemas.xmlsoap.org/soap/envelope/";
	public static final String SRWNamespace    = "http://www.loc.gov/zing/srw/";
	public static final String diagNamespace   = "http://www.loc.gov/zing/srw/diagnostic/";

	static XPathContext sruContext = new XPathContext();

	private void fillContext(){
		sruContext.addNamespace("SOAP-ENV", SOAPNamespace);
		sruContext.addNamespace("srw", SRWNamespace);
		sruContext.addNamespace("diag", diagNamespace);
	}

	public static Properties sruDefaultProps = new Properties();
	public static Properties locatorConfig = null;
	public static Properties httpConfig = null;

	static public Logger logger;

	static private String [] prefixPaths;
	static String myName;
	static String myVersion;

	private void setSRUDefaults(){
		sruDefaultProps.setProperty("operation", "searchRetrieve");
		sruDefaultProps.setProperty("version", "1.1");
		if ( locatorConfig.getProperty("sruMaximumRecords") == null )
			sruDefaultProps.setProperty("maximumRecords", "5");
		else
			sruDefaultProps.setProperty( "maximumRecords", 
					locatorConfig.getProperty("sruMaximumRecords"));
	}

	static public File getFile( String name ){
		return getFile( prefixPaths, name );
	}

	static	private File getFile( String[] prefixes, String name ){
		try {
			int i=0;
			while ( i < prefixes.length ){
				String fullName = prefixes[i] + File.separator + name;
				File file = new File( fullName );
				if ( file.exists() ){
					return file;
				}
				++i;
			}
		}
		catch ( Exception e ){
		}
		return null;
	}

	private Properties readConfig( String name ){
		Properties result = null;
		try {
			File configFile = getFile( prefixPaths, name );
			if ( configFile != null ){
				FileInputStream is = new FileInputStream(configFile);
				result = new Properties();
				result.load( is );
			}		
		}
		catch ( Exception e ){
			logger.error("readConfig problem: "+e.getMessage() );
		}
		return result;
	}	

	public void init() throws ServletException{
		myName = getServletContext().getInitParameter("artifactId" );
		myVersion = getServletContext().getInitParameter("version" );
		String prefixPath = getServletContext().getInitParameter("prefixPath" );
		if ( prefixPath.equals("") )
			throw new ServletException("no prefixPath in web.xmlfound" );
		prefixPaths = prefixPath.split(":");
		if ( prefixPaths.length == 0 ){
			throw new ServletException("no prefix paths for configuration found" );
		}
		String configName = null;
		File l4j = getFile( "log4j.properties");
		if ( l4j != null){
			configName = l4j.getAbsolutePath();
			PropertyConfigurator.configure( configName );
		}
		else {
			BasicConfigurator.configure();
		}
		logger = Logger.getLogger(Locator.class);
		logger.info("initialize Locator, using config from PATH=" + prefixPath );
		if ( configName != null )
			logger.debug("log4j configured from " + configName );
		else
			logger.warn("couldn't get log4j.properties. logging to System.out");
		fillContext();
		logger.info("started " + myName + "-" + myVersion );
	}

	static boolean doMeresco = false;

	enum serveAction {unknownAct, infoAct, explainAct, lookupAct };

	public void doGet( HttpServletRequest request, 
			HttpServletResponse response ) throws ServletException, IOException {

		OutputStream os = response.getOutputStream();
		PrintStream out = new PrintStream( os );
		String fileName = "locator.cfg";
		locatorConfig = readConfig( fileName );
		if ( locatorConfig == null ){
			out.println("unable to read configfile " + fileName );
			logger.error("reading configfile '" + fileName + "' failed");
			return;
		}
		logger.debug("config:" + locatorConfig.toString());
		String httpSettings = locatorConfig.getProperty("httpDefaults");
		if ( httpSettings != null ){
			fileName = httpSettings;
			httpConfig = readConfig(fileName);
			if ( httpConfig == null ){
				logger.warn("reading http configfile '" + fileName  + "' failed. (using defaults)");
			}
			else
				logger.debug("HTTP config:" + httpConfig.toString());
		}
		setSRUDefaults();
		response.setContentType("text/XML");
		String action = request.getParameter( "action" ); 	 
		String doM = locatorConfig.getProperty("doMeresco", "false");
		if ( doM.compareToIgnoreCase("true") == 0 ){
			doMeresco = true;
		}
		else
			doMeresco = false;
		Document doc = null;
		serveAction act = serveAction.lookupAct; 	
		if ( action != null ){ 	 
			logger.debug("special_action="+action); 	 
			if ( action.equals("info") ) 	 
				act = serveAction.infoAct; 	 
			else if ( action.equals("explain") ) 	 
				act = serveAction.explainAct; 	 
			else if ( action.equals("meresco") ){
				doMeresco = true;
			}
			else if ( action.equals("tucat") ){
				doMeresco = false;
			}
			else  {
				doc = errorDoc( "invalid action:" + action );
			}
		}
		if ( doc == null ){
			if ( act == serveAction.explainAct ){
				doc = explainDoc();
			}
			else if ( act == serveAction.infoAct ){
				ServeData sD = new ServeData();
				if ( sD.parseQuery( request ) ){
					logger.debug( "parsed query: " + sD );
					doc = tryToServeDoc( sD );
				}
				else
					doc = errorDoc("invalid parameters");
			}
			else {
				QueryData pQ = new QueryData();
				if ( pQ.parseQuery( request ) ){
					logger.debug( "parsed query: " + pQ );
					if ( pQ.genre.equals("article") || pQ.genre.equals("journal"))
						doc = tryToLocateArticle( pQ );
					else
						doc = tryToLocateBook( pQ );
				}
				else
					doc = errorDoc("invalid parameters");
			}
		}
		out.println( doc.toXML() );		
	}

		static DocServer dS = null;

		private Document tryToServeDoc( ServeData sD ) {
			if ( dS == null)
				dS = new DocServer();
			Document doc = dS.refresh( sD );
			if ( doc == null )
				doc = dS.serverDoc( sD );
			return doc;
		}

		private Document tryToLocateArticle( QueryData pQ ){
			List<LinkInfo> linkInfo = null;
			try {
				LinkDbConnection linkDb = new LinkDbConnection();
				linkInfo = linkDb.query( pQ );
			}
			catch ( LocatorException e ){
				logger.warn(e.getMessage());
				return errorDoc( 0, "linkDb failed", e.getMessage() );
			}
			if ( linkInfo != null ){
				for ( Iterator<LinkInfo> it=linkInfo.iterator(); it.hasNext(); ){
					LinkInfo li = it.next();
					try{
						if ( !li.inRange( pQ )){
							it.remove();
						}
						else if ( !li.applyMethod( pQ ) ){
							it.remove();
						}
					}
					catch (LocatorException e ){
						return errorDoc( 0, "threshold failed", e.getMessage() );
					}
				}
			}
			List<LinkInfo> tuInfo = null;
			if ( linkInfo == null || linkInfo.isEmpty() ){
				if ( doMeresco ){
					logger.debug("no valid linkdb results, try meresco");
					try {
						MerescoConnection bzv = new MerescoConnection();
						tuInfo = bzv.query( pQ );
					} catch (LocatorException e) {
						return errorDoc( 0, "meresco failed", e.getMessage() );
					}				
				}
				else {
					logger.debug("no valid linkdb results, try tucat");
					try {
						TucatConnection tucat = new TucatConnection();
						tuInfo = tucat.query( pQ );
					} catch (LocatorException e) {
						return errorDoc( 0, "tucat failed", e.getMessage() );
					}
				}
			}
			if ( tuInfo != null ){
				for ( Iterator<LinkInfo> it=tuInfo.iterator(); it.hasNext(); ){
					LinkInfo li = it.next();
					if ( !li.applyMethod( pQ ) ){
						it.remove();
					}
				}
				if (linkInfo != null )
					linkInfo.addAll( tuInfo );
				else
					linkInfo = tuInfo;
			}
			boolean suc6 = false;
			if ( linkInfo != null ){
				for ( Iterator<LinkInfo> it=linkInfo.iterator(); it.hasNext() && !suc6; ){
					LinkInfo li = it.next();
					suc6 = li.diagnostics.isEmpty();
				}
			}
			prune( linkInfo );
			if ( !suc6 )
				return IBLdoc( pQ, linkInfo );
			else
				return toDoc(linkInfo );
		}

		private Document tryToLocateBook( QueryData pQ ){
			if ( pQ.database == null ){
				return errorDoc( "missing 'database=' in query");
			}
			else {
				List<LinkInfo> linkInfo = null;
				if ( doMeresco ){
					try {
						MerescoConnection bzv = new MerescoConnection();
						linkInfo = bzv.query( pQ );
					} catch (LocatorException e) {
						return errorDoc( 0, "meresco failed", e.getMessage() );
					}				
				}
				else {
					try {
						TucatConnection tucat = new TucatConnection();
						linkInfo = tucat.query( pQ );
					} catch (LocatorException e) {
						return errorDoc( 0, "tucat failed", e.getMessage() );
					}
				}
				if ( linkInfo != null ){
					for ( Iterator<LinkInfo> it=linkInfo.iterator(); it.hasNext(); ){
						LinkInfo li = it.next();
						if ( !li.applyMethod( pQ ) ){
							it.remove();
						}
					}
				}
				if ( linkInfo != null && !linkInfo.isEmpty() ){
					prune( linkInfo );
					return toDoc(linkInfo );
				}
				else {
					return IBLdoc( pQ, null );		
				}
			}
		}

		private void prune(List<LinkInfo> linkInfo) {
			//
			// prune doubles.
			// For now only 1 url in tucat records
			//
			if ( linkInfo != null ){
				String urlStore = "";
				for ( Iterator<LinkInfo> it=linkInfo.iterator(); it.hasNext(); ){
					LinkInfo li = it.next();
					if ( li.getClass() == TucatInfo.class ){
						TucatInfo tu = (TucatInfo)li;
						if ( tu.action.equals("url") )
							if ( tu.generatedUrl.equals(urlStore) )
								it.remove();
							else if ( urlStore.equals("") )
								urlStore = tu.generatedUrl;
					}
				}
			}
		}	

		private Document IBLdoc( QueryData qD, List<LinkInfo> linkInfoList ) {
			IllRequestor illRequestor = new IllRequestor();
			LinkInfo li;
			if ( linkInfoList == null || linkInfoList.size() ==0  )
				li = new LinkInfo(); //empty stub
			else
				li = linkInfoList.get(0);
			Document doc = illRequestor.getDocument( qD, li ); 
			return doc;
		}

		public Document toDoc( List<LinkInfo> linkInfo ){
			Element root = new Element("records");
			Document doc = new Document( root );
			if ( linkInfo != null ){
				for ( int i=0; i< linkInfo.size(); ++i ){
					Node liNode = linkInfo.get(i).toRecordNode();
					root.appendChild( liNode );
				}
			}
			return doc;
		}

		static public int extractHits( Document doc ){
			Nodes numOfRecords = doc.query("//srw:numberOfRecords", sruContext);
			if( numOfRecords.size() > 0 ){
				Element numHitsElement = (Element) numOfRecords.get(0);
				String hitString = numHitsElement.getValue();
				int hits=Integer.parseInt(hitString);
				return hits;
			}
			return 0;
		}

		public static nu.xom.Document httpGetDocument( String url, Properties args )
		throws UnsupportedEncodingException, MalformedURLException, IOException, ValidityException, ParsingException {
			nu.xom.Document responseDocument = null;
			InputStream is = HttpGetInputStream(url, args);
			Builder builder = new Builder();
			responseDocument =  builder.build(is);
			return responseDocument;
		}

		public static nu.xom.Document httpGetDocument( String url )
		throws UnsupportedEncodingException, MalformedURLException, IOException, ValidityException, ParsingException {
			return httpGetDocument( url, null ); 
		}

		/**
		 * Perform a HTTP GET request and return InputStream.
		 * Based on: Jason Hunter & William Crawford, Java Servlet Programming, 2nd Edition,
		 * pp 316-319, O'Reilly
		 * 
		 * @param url URL of web service
		 * @param args key/value pairs
		 * @return InputStream with response
		 */

		public static String createUrlString( String base, Properties attVals ) throws UnsupportedEncodingException{
			String url = base;
			String argString = "";
			if ( attVals != null) {
				argString = "?" + toEncodedString(attVals);
				url += argString;
			}
			return url;
		}

		public static InputStream HttpGetInputStream( String url, Properties props ) 
		throws UnsupportedEncodingException,
		MalformedURLException,
		IOException {
			URL fullUrl = new URL( url );
			URLConnection con = fullUrl.openConnection();
			if ( props != null ){
				String useCache = props.getProperty("useCache", "false");
				if ( useCache.equalsIgnoreCase("true") )
					con.setUseCaches(true);
				else
					con.setUseCaches(false);
//				logger.debug("HTTP set useCache=" + useCache );
				String keepAlive = props.getProperty("Keep-Alive", "300");
				con.setRequestProperty("Keep-Alive", keepAlive  );
//				logger.debug("HTTP set Keep-Alive=" + keepAlive );
				String timeout = props.getProperty("urlTimeout", "10000"); // take max 10 seconds, or what is configured
				con.setReadTimeout( Integer.parseInt(timeout) );
//				logger.debug("HTTP set timeout=" + timeout );
			}
			return con.getInputStream();
		}

		/**
		 * Transforms key/value pairs into query string
		 * 
		 * @param args Series of key/value pairs
		 * @return String like 'a=1&b=2', without question mark 
		 */

		private static String toEncodedString(Properties args)
		throws UnsupportedEncodingException {
			StringBuffer buf = new StringBuffer();
			Enumeration names = args.propertyNames();
			while (names.hasMoreElements()) {
				String name = (String) names.nextElement();
				String value = args.getProperty(name);
				buf.append(URLEncoder.encode(name, "ISO-8859-1") + "=" + URLEncoder.encode(value, "ISO-8859-1"));
				if (names.hasMoreElements())
					buf.append("&");
			}

			return buf.toString();
		}


		public static Document explainDoc( ) {
			Element explainNode = new Element("explain");
			Element versionNode = new Element("version");
			Text txt = new Text( myVersion );
			versionNode.appendChild( txt );
			Element artifactNode = new Element("artifactId");
			txt = new Text( myName );
			artifactNode.appendChild( txt );
			explainNode.appendChild( artifactNode);
			explainNode.appendChild( versionNode);
			return new Document( explainNode );
		}

		public static Document errorDoc( int code, String mess, String addI ) {
			Diagnostic diag = new Diagnostic( code, mess, addI );
			Element diagNode = diag.toElement();
			return new Document( diagNode );
		}

		public static Document errorDoc( String mess ) {
			return errorDoc( 1, mess, "" );
		}

		public Document notFoundPage( String what ) {
			return errorDoc( 2, "no hits", what );
		}	  


		static final long serialVersionUID=1;
	}
